// ==============================================================
//
// cygnss
// Control module for cygnss vessel class
//
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>

// ==============================================================
// Some vessel parameters
// ==============================================================



const double DRYMASS = 28;
const double SOLP_SPEED = 0.025;

static TOUCHDOWNVTX tdvtx_geardown[3] = {
	{_V(0, -0.1, 2), 1e4, 1e4, 0.5, 0.5},
	{_V(-1.732, -0.1, -1), 1e4, 1e4, 0.5, 0.5},
	{_V( 1.732, -0.1, -1), 1e4, 1e4, 0.5, 0.5}
};

#define GREEN RGB(125, 255, 125)


// ==============================================================
// Shuttle-PB class interface
// ==============================================================

class cygnss: public VESSEL3 {
public:
	cygnss (OBJHANDLE hVessel, int flightmodel);
	~cygnss ();
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);


private:


	ANIMATIONCOMPONENT_HANDLE Section, Section1;
	ATTACHMENTHANDLE hDispenser;

	void DefineAnimations (void);
	double step_time;
	int solp_status, asp_status;

	UINT anim_solp;
	double solp_proc;
	void DoNadir(void);
	void DoDrag(void);
	

	MGROUP_TRANSFORM *Solp1rot, *Solp2rot, *Solp3rot, *Solp4rot;

	
};

cygnss::cygnss (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
    

}

cygnss::~cygnss ()
{
	delete Solp1rot;
	delete Solp2rot;
	delete Solp3rot;
	delete Solp4rot;
	
}

void cygnss::DefineAnimations (void)
{
// 1. SOLP

	// x pos
	anim_solp = CreateAnimation (0);
	static UINT Solp1Grp[1] = {5};
	Solp1rot = new MGROUP_ROTATE(0, Solp1Grp, 1, _V(0.255, 0.032, 0), _V(0,0,1), (float)(-180*RAD));
	Section = AddAnimationComponent (anim_solp, 0.5, 1, Solp1rot);


	static UINT Solp2Grp[1] = {3};
	Solp2rot = new MGROUP_ROTATE(0, Solp2Grp, 1, _V(0.055, 0.035, 0), _V(0,0,1), (float)(180*RAD));
	AddAnimationComponent (anim_solp, 0.5, 1, Solp2rot, Section);

	// x neg
	static UINT Solp3Grp[1] = {4};
	Solp3rot = new MGROUP_ROTATE(0, Solp3Grp, 1, _V(-0.255, 0.032, 0), _V(0,0,1), (float)(180*RAD));
	Section = AddAnimationComponent (anim_solp, 0.5, 1, Solp3rot);

	static UINT Solp4Grp[1] = {2};
	Solp4rot = new MGROUP_ROTATE(0, Solp4Grp, 1, _V(-0.055, 0.035, 0), _V(0,0,1), (float)(-180*RAD));
	AddAnimationComponent (anim_solp, 0.5, 1, Solp4rot, Section);

}


// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void cygnss::clbkSetClassCaps (FILEHANDLE cfg)
{
	step_time = 0;

	solp_status = 1;

	asp_status = 0;

	solp_proc = 0.0;

	DefineAnimations ();
	// physical vessel parameters
	SetSize (0.8);
	SetEmptyMass (DRYMASS);
	SetPMI(_V(0.04, 0.05, 0.02));
	SetCrossSections (_V(0.1,0.33,0.07));
	SetCW(5,5,10,40);
	SetRotDrag (_V(2,2,2));
	SetTouchdownPoints (tdvtx_geardown, 3);
	SetSurfaceFrictionCoeff (1,1);
	// camera parameters
	SetCameraOffset (_V(0,1.5,-1));

	SetMeshVisibilityMode(AddMesh("cygnss/cygnss"),MESHVIS_EXTERNAL);


	hDispenser = CreateAttachment(true,_V(0,-0.135,0), _V(0,-1,0), _V(0,0,1), "XS");
	
}



// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new cygnss (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (cygnss*)vessel;
}


void cygnss::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState(scn);

	sprintf (cbuf, "%d %0.4f", solp_status, solp_proc);
	oapiWriteScenario_string (scn, "SOLP", cbuf);

	
	sprintf (cbuf, "%d", asp_status);
	oapiWriteScenario_string (scn, "KROT", cbuf);


}
void cygnss::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "SOLP", 4)) {
			sscanf (line+4, "%d%lf", &solp_status, &solp_proc);
		} else if (!strnicmp (line, "KROT", 4)) {
			sscanf (line+4, "%d", &asp_status);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }

	SetAnimation(anim_solp, solp_proc);


}



int cygnss::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {



if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) 
  {	 
	switch (key)
	{
		

	} 
  }

  else if (KEYMOD_CONTROL (kstate)) 
  { 
	switch (key)
	{
		
	} 
  }

else { // unmodified keys
    switch (key) {


		case OAPI_KEY_K: // attitude pointing
		if (asp_status == 0)asp_status = 1; // nadir to earth
		else if (asp_status == 1)asp_status = 2; // nadir to velocity vector
		else if (asp_status == 2)asp_status = 3; // killrot
		else asp_status = 0; // off
		return 1;


	}
  }
  return 0;
}


void cygnss::clbkPreStep(double SimT, double SimDT, double MJD)
{

step_time = SimDT;///////
double simdt = SimDT;

if(solp_status == 0)
{
if (GetAttachmentStatus(hDispenser) == NULL)solp_status = 1;
}

if(asp_status == 1) DoNadir();
else if(asp_status == 2) DoDrag();
else if(asp_status == 3)		// killrot
{
VECTOR3 avel = _V(0,0,0);
GetAngularVel(avel);
double drot = RAD*2*simdt;

if (avel.x > -drot && avel.x < drot) avel.x = 0;
else if (avel.x > drot) avel.x = avel.x - drot;
else if (avel.x < -drot) avel.x = avel.x + drot;

if (avel.y > -drot && avel.y < drot) avel.y = 0;
else if (avel.y > drot) avel.y = avel.y - drot;
else if (avel.y < -drot) avel.y = avel.y + drot;

if (avel.z > -drot && avel.z < drot) avel.z = 0;
else if (avel.z > drot) avel.z = avel.z - drot;
else if (avel.z < -drot) avel.z = avel.z + drot;

SetAngularVel(avel);
}



////////////////// ANIMATIONS ///////////////

// solp
if (solp_status == 1)
{
double da = SimDT * SOLP_SPEED;

solp_proc = solp_proc+da;
if (solp_proc > 1)
	{
	solp_proc = 1;
	solp_status = 2;
	asp_status = 1;
	}
SetAnimation(anim_solp, solp_proc);
}

else if (solp_status == 2)
{
SetAnimation(anim_solp, 1);
}


}

void cygnss::clbkPostStep (double simt, double simdt, double mjd)
{
	
}




bool cygnss::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);

	//int chw = 16;

	
	char test[256];
	strcpy(test,"abcdefghijklmnopqrstuvwxyz");
	int it = skp->GetTextWidth(test, 26);
	it = (int)(it/26);
	int chw = it;

	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	/*int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	int w5 = (int)(264*d);
	int w6 = (int)(280*d);
	int w7 = (int)(296*d);
	int w8 = (int)(312*d);
	int w9 = (int)(328*d);
	int w10 = (int)(344*d);
	int w11 = (int)(360*d);
	int w12 = (int)(376*d);
	int w13 = (int)(392*d);
	int w14 = (int)(408*d);
	int w15 = (int)(424*d);*/

	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Att.Ctrl.Mode = K");
	skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
	}

	if (asp_status == 0)
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Attitude Ctrl.Off");
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}
	else if (asp_status == 1)
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Nadir Pointing Active");
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}

	else if (asp_status == 2)
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Max Drag Active");
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}

	else if (asp_status == 3)
	{
	skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"KillRot Active");
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}




	return true;
}


void cygnss::DoNadir()
{
	VECTOR3 avel = _V(0,0,0);
	GetAngularVel(avel);
	double dt = oapiGetSimStep();
	double max_acc = 2*RAD;
	double drot = RAD*2*dt;

	char cbuf[256]="Sun";
	OBJHANDLE hRef = oapiGetGbodyByName (cbuf);
	VECTOR3 sun_pos_global = _V(0,0,0);
	oapiGetGlobalPos(hRef, &sun_pos_global);

	char abuf[256]="Earth";
	OBJHANDLE eRef = oapiGetGbodyByName (abuf);
	VECTOR3 earth_pos_global = _V(0,0,0);
	oapiGetGlobalPos(eRef, &earth_pos_global);


	VECTOR3 sun_dir_local = _V(0,0,0);
	Global2Local(sun_pos_global,sun_dir_local);
	normalise(sun_dir_local);
	sun_dir_local = -sun_dir_local;

	VECTOR3 vel_vect_local = _V(0,0,0);
	GetAirspeedVector(FRAME_LOCAL, vel_vect_local);
	normalise(vel_vect_local);

	sun_dir_local = vel_vect_local;

	VECTOR3 earth_dir_local = _V(0,0,0);
	Global2Local(earth_pos_global,earth_dir_local);
	normalise(earth_dir_local);

	VECTOR3 norm_ess_local = _V(0,0,0);
	norm_ess_local = crossp(sun_dir_local, earth_dir_local);
	normalise(norm_ess_local);

	VECTOR3 zfwd_local = _V(0,0,0);
	zfwd_local = crossp(earth_dir_local, norm_ess_local);
	normalise(zfwd_local);

	double xaxis_angle = atan2(earth_dir_local.x,-earth_dir_local.y);
	if (xaxis_angle < 0)xaxis_angle = xaxis_angle + (2*PI);

	double zaxis_angle = atan2(sun_dir_local.x,sun_dir_local.z);
	if (zaxis_angle < 0)zaxis_angle = zaxis_angle + (2*PI);

	double yaxis_angle = atan2(zfwd_local.y, zfwd_local.z);
	if (yaxis_angle < 0)yaxis_angle = yaxis_angle + (2*PI);

	double max_dec_ang = 0;
	double max_rot = 2*RAD;


	if (xaxis_angle < PI)
	{
	max_dec_ang = (avel.z/max_acc)*avel.z*0.5;
	if (max_dec_ang > xaxis_angle)
	{	
		if (avel.z < 0)avel.z = avel.z+drot;
		else avel.z = avel.z-drot;
	}
	else if (avel.z > -max_rot)avel.z = avel.z-drot;
	else avel.z = avel.z+drot;
		
	}
	else
	{
	max_dec_ang = (avel.z/max_acc)*avel.z*0.5;
	if (max_dec_ang > ((2*PI)-xaxis_angle))
		{	
		if (avel.z > 0)avel.z = avel.z-drot;
		else avel.z = avel.z+drot;
		}
	else if (avel.z < max_rot)avel.z = avel.z+drot;
	else avel.z = avel.z-drot;
	}



	if (yaxis_angle < PI)
	{
	max_dec_ang = (avel.x/max_acc)*avel.x*0.5;
	if (max_dec_ang > yaxis_angle)
		{	
		if (avel.x < 0)avel.x = avel.x+drot;
		else avel.x = avel.x-drot;
		}
	else if (avel.x < max_rot)avel.x = avel.x+drot;
	else avel.x = avel.x-drot;
	}
	else
	{
	max_dec_ang = (avel.x/max_acc)*avel.x*0.5;
	if (max_dec_ang > ((2*PI)-yaxis_angle))
		{	
		if (avel.x > 0)avel.x = avel.x-drot;
		else avel.x = avel.x+drot;
		}
	else if (avel.x > -max_rot)avel.x = avel.x-drot;
	else avel.x = avel.x+drot;
	}



	if (zaxis_angle < PI)
	{
	max_dec_ang = (avel.y/max_acc)*avel.y*0.5;
	if (max_dec_ang > zaxis_angle)
		{	
		if (avel.y < 0)avel.y = avel.y+drot;
		else avel.y = avel.y-drot;
		}
	else if (avel.y > -max_rot)avel.y = avel.y-drot;
	else avel.y = avel.y+drot;
	}
	else
	{
	max_dec_ang = (avel.y/max_acc)*avel.y*0.5;
	if (max_dec_ang > ((2*PI)-zaxis_angle))
		{	
		if (avel.y > 0)avel.y = avel.y-drot;
		else avel.y = avel.y+drot;
		}
	else if (avel.y < max_rot)avel.y = avel.y+drot;
	else avel.y = avel.y-drot;
	}

	SetAngularVel(avel);
	double ang_dif_x = xaxis_angle;
	if (ang_dif_x > PI)ang_dif_x = (2*PI)-xaxis_angle;

	double ang_dif_y = yaxis_angle;
	if (ang_dif_y > PI)ang_dif_y = (2*PI)-yaxis_angle;

	double ang_dif_z = zaxis_angle;
	if (ang_dif_z > PI)ang_dif_z = (2*PI)-zaxis_angle;

	
	if ((drot*5) > ang_dif_x && (drot*5) > ang_dif_y && (drot*5) > ang_dif_z)
	{


	VECTOR3 yup = _V(0,0,0);
	VECTOR3 zforward = _V(0,0,0);
	VECTOR3 xright = _V(0,0,0);

	GlobalRot(norm_ess_local, xright);
	GlobalRot(-earth_dir_local, yup);
	GlobalRot(zfwd_local,zforward);
	
	normalise(zforward);
	normalise(yup);
	normalise(xright);
	
	double gamma = atan2( xright.x, yup.x);
	double beta = -1*(asin(zforward.x));
	double alpha = atan2(zforward.y, zforward.z);
	SetGlobalOrientation (_V(alpha, beta, (PI/2)-gamma));
	SetAngularVel (_V(0,0,0));

	}
	//sprintf(oapiDebugString(), "yup %0.4f %0.4f %0.4f", yup.x, yup.y, yup.z);
}

void cygnss::DoDrag()
{
	VECTOR3 avel = _V(0,0,0);
	GetAngularVel(avel);
	double dt = oapiGetSimStep();
	double max_acc = 2*RAD;
	double drot = RAD*2*dt;

	char cbuf[256]="Sun";
	OBJHANDLE hRef = oapiGetGbodyByName (cbuf);
	VECTOR3 sun_pos_global = _V(0,0,0);
	oapiGetGlobalPos(hRef, &sun_pos_global);

	char abuf[256]="Earth";
	OBJHANDLE eRef = oapiGetGbodyByName (abuf);
	VECTOR3 earth_pos_global = _V(0,0,0);
	oapiGetGlobalPos(eRef, &earth_pos_global);


	VECTOR3 sun_dir_local = _V(0,0,0);
	Global2Local(sun_pos_global,sun_dir_local);
	normalise(sun_dir_local);
	sun_dir_local = -sun_dir_local;

	VECTOR3 vel_vect_local = _V(0,0,0);
	GetAirspeedVector(FRAME_LOCAL, vel_vect_local);
	normalise(vel_vect_local);

	//sun_dir_local = vel_vect_local;

	VECTOR3 earth_dir_local = _V(0,0,0);
	Global2Local(earth_pos_global,earth_dir_local);
	normalise(earth_dir_local);

	sun_dir_local = earth_dir_local;
	earth_dir_local = vel_vect_local;

	VECTOR3 norm_ess_local = _V(0,0,0);
	norm_ess_local = crossp(sun_dir_local, earth_dir_local);
	normalise(norm_ess_local);

	VECTOR3 zfwd_local = _V(0,0,0);
	zfwd_local = crossp(earth_dir_local, norm_ess_local);
	normalise(zfwd_local);

	double xaxis_angle = atan2(earth_dir_local.x,-earth_dir_local.y);
	if (xaxis_angle < 0)xaxis_angle = xaxis_angle + (2*PI);

	double zaxis_angle = atan2(sun_dir_local.x,sun_dir_local.z);
	if (zaxis_angle < 0)zaxis_angle = zaxis_angle + (2*PI);

	double yaxis_angle = atan2(zfwd_local.y, zfwd_local.z);
	if (yaxis_angle < 0)yaxis_angle = yaxis_angle + (2*PI);

	double max_dec_ang = 0;
	double max_rot = 2*RAD;


	if (xaxis_angle < PI)
	{
	max_dec_ang = (avel.z/max_acc)*avel.z*0.5;
	if (max_dec_ang > xaxis_angle)
	{	
		if (avel.z < 0)avel.z = avel.z+drot;
		else avel.z = avel.z-drot;
	}
	else if (avel.z > -max_rot)avel.z = avel.z-drot;
	else avel.z = avel.z+drot;
		
	}
	else
	{
	max_dec_ang = (avel.z/max_acc)*avel.z*0.5;
	if (max_dec_ang > ((2*PI)-xaxis_angle))
		{	
		if (avel.z > 0)avel.z = avel.z-drot;
		else avel.z = avel.z+drot;
		}
	else if (avel.z < max_rot)avel.z = avel.z+drot;
	else avel.z = avel.z-drot;
	}



	if (yaxis_angle < PI)
	{
	max_dec_ang = (avel.x/max_acc)*avel.x*0.5;
	if (max_dec_ang > yaxis_angle)
		{	
		if (avel.x < 0)avel.x = avel.x+drot;
		else avel.x = avel.x-drot;
		}
	else if (avel.x < max_rot)avel.x = avel.x+drot;
	else avel.x = avel.x-drot;
	}
	else
	{
	max_dec_ang = (avel.x/max_acc)*avel.x*0.5;
	if (max_dec_ang > ((2*PI)-yaxis_angle))
		{	
		if (avel.x > 0)avel.x = avel.x-drot;
		else avel.x = avel.x+drot;
		}
	else if (avel.x > -max_rot)avel.x = avel.x-drot;
	else avel.x = avel.x+drot;
	}



	if (zaxis_angle < PI)
	{
	max_dec_ang = (avel.y/max_acc)*avel.y*0.5;
	if (max_dec_ang > zaxis_angle)
		{	
		if (avel.y < 0)avel.y = avel.y+drot;
		else avel.y = avel.y-drot;
		}
	else if (avel.y > -max_rot)avel.y = avel.y-drot;
	else avel.y = avel.y+drot;
	}
	else
	{
	max_dec_ang = (avel.y/max_acc)*avel.y*0.5;
	if (max_dec_ang > ((2*PI)-zaxis_angle))
		{	
		if (avel.y > 0)avel.y = avel.y-drot;
		else avel.y = avel.y+drot;
		}
	else if (avel.y < max_rot)avel.y = avel.y+drot;
	else avel.y = avel.y-drot;
	}

	SetAngularVel(avel);
	double ang_dif_x = xaxis_angle;
	if (ang_dif_x > PI)ang_dif_x = (2*PI)-xaxis_angle;

	double ang_dif_y = yaxis_angle;
	if (ang_dif_y > PI)ang_dif_y = (2*PI)-yaxis_angle;

	double ang_dif_z = zaxis_angle;
	if (ang_dif_z > PI)ang_dif_z = (2*PI)-zaxis_angle;

	
	if ((drot*5) > ang_dif_x && (drot*5) > ang_dif_y && (drot*5) > ang_dif_z)
	{


	VECTOR3 yup = _V(0,0,0);
	VECTOR3 zforward = _V(0,0,0);
	VECTOR3 xright = _V(0,0,0);

	GlobalRot(norm_ess_local, xright);
	GlobalRot(-earth_dir_local, yup);
	GlobalRot(zfwd_local,zforward);
	
	normalise(zforward);
	normalise(yup);
	normalise(xright);
	
	double gamma = atan2( xright.x, yup.x);
	double beta = -1*(asin(zforward.x));
	double alpha = atan2(zforward.y, zforward.z);
	SetGlobalOrientation (_V(alpha, beta, (PI/2)-gamma));
	SetAngularVel (_V(0,0,0));

	}
}