// ==============================================================
//
// cygnss_dispenser.cpp
// Control module for cygnss_dispenser vessel class
//
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>

// ==============================================================
// Some vessel parameters
// ==============================================================


const double DRYMASS = 100;

#define GREEN RGB(125, 255, 125)




// ==============================================================
// Shuttle-PB class interface
// ==============================================================

class cygnss_dispenser: public VESSEL3 {
public:
	cygnss_dispenser (OBJHANDLE hVessel, int flightmodel);
	~cygnss_dispenser ();
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	void clbkPostStep (double simt, double simdt, double mjd);
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	void Jettison(void);


private:

	int npl_status;
	ATTACHMENTHANDLE cyg_attach[8];
	double deploy_timer;
	
};

cygnss_dispenser::cygnss_dispenser (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
    

}

cygnss_dispenser::~cygnss_dispenser ()
{

}



// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void cygnss_dispenser::clbkSetClassCaps (FILEHANDLE cfg)
{
	npl_status = 0;
	deploy_timer = 10;

	// physical vessel parameters
	SetSize (2);
	SetEmptyMass (DRYMASS);
	SetPMI(_V(2,2,0.75));
	SetCrossSections (_V(5,5,3));
	SetCW(1,1,1,1);
	SetRotDrag (_V(2,2,1));
	SetTouchdownPoints ( _V(0,-1, 5), _V(-2,-1,-5), _V( 2,-1,-5));

	// camera parameters
	SetCameraOffset (_V(0,2,-2));

	SetMeshVisibilityMode(AddMesh("cygnss/cygnss_dispenser"),MESHVIS_ALWAYS);


	// propellant resources
	

	// ***************** thruster definitions *******************
	

	///////////////////////////////////////////////////////
	/////AirFoils


	 cyg_attach[0] = CreateAttachment (false, _V( 0.24, 0, 1.3), _V( 1,0,0), _V(0,0,1), "PL1");
	 cyg_attach[1] = CreateAttachment (false, _V(-0.24, 0, 1.3), _V(-1,0,0), _V(0,0,1), "PL2");
	 cyg_attach[2] = CreateAttachment (false, _V( 0, 0.24, 1.3), _V(0, 1,0), _V(0,0,1), "PL3");
	 cyg_attach[3] = CreateAttachment (false, _V( 0,-0.24, 1.3), _V(0,-1,0), _V(0,0,1), "PL4");

	 cyg_attach[4] = CreateAttachment (false, _V( 0.17, 0.17, 0.616), _V( 0.7071, 0.7071,0), _V(0,0,1), "PL5");
	 cyg_attach[5] = CreateAttachment (false, _V(-0.17,-0.17, 0.616), _V(-0.7071,-0.7071,0), _V(0,0,1), "PL6");
	 cyg_attach[6] = CreateAttachment (false, _V( 0.17,-0.17, 0.616), _V( 0.7071,-0.7071,0), _V(0,0,1), "PL7");
	 cyg_attach[7] = CreateAttachment (false, _V(-0.17, 0.17, 0.616), _V(-0.7071, 0.7071,0), _V(0,0,1), "PL8");



}



// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new cygnss_dispenser (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (cygnss_dispenser*)vessel;
}


void cygnss_dispenser::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState(scn);

	//sprintf (cbuf, "%d %0.4f %d %0.4f", solp_status, solp_proc, st_status, st_proc);
	//oapiWriteScenario_string (scn, "SOLP", cbuf);

	//sprintf (cbuf, "%d %0.4f", mast_status, mast_proc);
	//oapiWriteScenario_string (scn, "LDMA", cbuf);
	
	sprintf (cbuf, "%d", npl_status);
	oapiWriteScenario_string (scn, "NPLS", cbuf);


}
void cygnss_dispenser::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "NPLS", 4)) {
			sscanf (line+4, "%d", &npl_status);
		//} else if (!strnicmp (line, "LDMA", 4)) {
		//	sscanf (line+4, "%d%lf", &mast_status, &mast_proc);
		//} else if (!strnicmp (line, "KROT", 4)) {
		//	sscanf (line+4, "%d%d",&kr_status, &asp_status);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }




}



int cygnss_dispenser::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {



if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) 
  {	 
	switch (key)
	{
		

	} 
  }

  else if (KEYMOD_CONTROL (kstate)) 
  { 
	switch (key)
	{
		
	} 
  }

else { // unmodified keys
    switch (key) {


		case OAPI_KEY_J: // jettison
		if (npl_status == 0) npl_status = 1;
		return 1;


	}
  }
  return 0;
}



void cygnss_dispenser::clbkPreStep(double SimT, double SimDT, double MJD)
{
	double pld_mass =0.0;

	for (DWORD j = 0; j < 8; j++)
	{
	OBJHANDLE hOg = GetAttachmentStatus(cyg_attach[j]);
	if (hOg == NULL) continue;
	else pld_mass = pld_mass + oapiGetMass(hOg);
	}

	double total_mass = DRYMASS + pld_mass;
	SetEmptyMass(total_mass);

	/////////
	if (npl_status > 0 && npl_status < 8)
	{
		deploy_timer = deploy_timer + SimDT;
		if (deploy_timer > 10)
		{
		deploy_timer = 0;
		Jettison();
		}
	}
//sprintf(oapiDebugString(), "%d %0.4f %0.4f", (sa1_stbd_elb_sts + sa1_stbd_scale_sts + sa1_port_elb_sts + sa1_port_scale_sts),sa1_port_rot_proc,sa1_stbd_rot_proc);
}

void cygnss_dispenser::clbkPostStep (double simt, double simdt, double mjd)
{
	
}

void cygnss_dispenser::Jettison(void)
{
DetachChild(cyg_attach[npl_status-1], 0.3);
npl_status = npl_status+1;
DetachChild(cyg_attach[npl_status-1], 0.3);
npl_status = npl_status+1;
}


bool cygnss_dispenser::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);

	//int chw = 16;

	
	char test[256];
	strcpy(test,"abcdefghijklmnopqrstuvwxyz");
	int it = skp->GetTextWidth(test, 26);
	it = (int)(it/26);
	int chw = it;

	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	int w5 = (int)(264*d);
	int w6 = (int)(280*d);
	int w7 = (int)(296*d);
	int w8 = (int)(312*d);
	int w9 = (int)(328*d);
	int w10 = (int)(344*d);
	int w11 = (int)(360*d);
	int w12 = (int)(376*d);
	int w13 = (int)(392*d);
	int w14 = (int)(408*d);
	int w15 = (int)(424*d);



	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"J = Start Jettison Sequence");
	skp->Text ((10+roxl), (wc+royl), abuf,strlen(abuf));
	}

	
	
	return true;
}
